from .basic_euclidean import eucl_dist
import math
import numpy as np

PI = math.pi
HPI = math.pi / 2


def ordered_mixed_distance(si, ei, sj, ej, siei, sjej, siei_norm_2, sjej_norm_2):
    siei_norm = math.sqrt(siei_norm_2)
    sjej_norm = math.sqrt(sjej_norm_2)
    sisj = sj - si
    siej = ej - si

    u1 = (sisj[0] * siei[0] + sisj[1] * siei[1]) / siei_norm_2
    u2 = (siej[0] * siei[0] + siej[1] * siei[1]) / siei_norm_2

    ps = si + u1 * siei
    pe = si + u2 * siei

    cos_theta = max(-1, min(1, (sjej[0] * siei[0] + sjej[1] * siei[1]) / (siei_norm * sjej_norm)))
    theta = math.acos(cos_theta)

    # perpendicular distance
    lpe1 = eucl_dist(sj, ps)
    lpe2 = eucl_dist(ej, pe)
    if lpe1 == 0 and lpe2 == 0:
        dped = 0
    else:
        dped = (lpe1 * lpe1 + lpe2 * lpe2) / (lpe1 + lpe2)

    # parallel_distance
    lpa1 = min(eucl_dist(si, ps), eucl_dist(ei, ps))
    lpa2 = min(eucl_dist(si, pe), eucl_dist(ei, pe))
    dpad = min(lpa1, lpa2)

    # angle_distance
    if 0 <= theta < HPI:
        dad = sjej_norm * math.sin(theta)
    elif HPI <= theta <= PI:
        dad = sjej_norm
    else:
        raise ValueError("WRONG THETA")

    fdist = (dped + dpad + dad) / 3

    return fdist


def mixed_distance(si, ei, sj, ej):
    siei = ei - si
    sjej = ej - sj

    siei_norm_2 = (siei[0] * siei[0]) + (siei[1] * siei[1])
    sjej_norm_2 = (sjej[0] * sjej[0]) + (sjej[1] * sjej[1])

    if sjej_norm_2 > siei_norm_2:
        md = ordered_mixed_distance(sj, ej, si, ei, sjej, siei, sjej_norm_2, siei_norm_2)
    else:
        md = ordered_mixed_distance(si, ei, sj, ej, siei, sjej, siei_norm_2, sjej_norm_2)

    return md


def segments_distance(traj_0, traj_1):
    n0 = len(traj_0)
    n1 = len(traj_1)
    M = np.zeros((n0 - 1, n1 - 1))
    for i in range(n0 - 1):
        for j in range(n1 - 1):
            M[i, j] = mixed_distance(traj_0[i], traj_0[i + 1], traj_1[j], traj_1[j + 1])
    return M



from setuptools import setup
from distutils.extension import Extension
from Cython.Distutils import build_ext
import numpy

ext_modules = [Extension("traj_dist.cydist.basic_geographical", ["traj_dist/cydist/basic_geographical.pyx"]),
               Extension("traj_dist.cydist.basic_euclidean", ["traj_dist/cydist/basic_euclidean.pyx"]),
               Extension("traj_dist.cydist.sspd", ["traj_dist/cydist/sspd.pyx"]),
               Extension("traj_dist.cydist.dtw", ["traj_dist/cydist/dtw.pyx"]),
               Extension("traj_dist.cydist.lcss", ["traj_dist/cydist/lcss.pyx"]),
               Extension("traj_dist.cydist.hausdorff", ["traj_dist/cydist/hausdorff.pyx"]),
               Extension("traj_dist.cydist.discret_frechet", ["traj_dist/cydist/discret_frechet.pyx"]),
               Extension("traj_dist.cydist.frechet", ["traj_dist/cydist/frechet.pyx"]),
               Extension("traj_dist.cydist.segment_distance", ["traj_dist/cydist/segment_distance.pyx"]),
               Extension("traj_dist.cydist.sowd", ["traj_dist/cydist/sowd.pyx"]),
               Extension("traj_dist.cydist.erp", ["traj_dist/cydist/erp.pyx"]),
               Extension("traj_dist.cydist.edr", ["traj_dist/cydist/edr.pyx"])]

setup(
    name="traj_dist",
    version="1.13",
    license="MIT",
    author="Brendan Guillouet",
    author_email="brendan.guillouet@gmail.com",
    cmdclass={'build_ext': build_ext},
    ext_modules=ext_modules,
    include_dirs=[numpy.get_include()],
    install_requires=["numpy>=1.14.0", "Cython>=0.27.3", "Shapely>=1.6.4", "geohash2==1.1", 'pandas>=0.20.3',
                      'scipy>=0.19.1'],
    setup_requires=["Cython>=0.27.3"],
    description="Distance to compare 2D-trajectories in Python/Cython",
    download_url = 'https://github.com/user/reponame/archive/1.13.tar.gz',    # I explain this later on
    keywords = ['trajectory',"distance","haversine"],
    packages=["traj_dist", "traj_dist.cydist", "traj_dist.pydist"],
    url = 'https://github.com/bguillouet/traj-dist',
)



STUFF = "Hi"

#cython: boundscheck=False
#cython: wraparound=False

cimport numpy as np
import numpy as np
from libc.stdlib cimport malloc, free

from libc.math cimport fmin
from basic_euclidean import c_eucl_dist
from cpython cimport bool
from numpy.math cimport INFINITY



cdef double _owd_grid_brut(np.ndarray[long,ndim=2] traj_cell_1, np.ndarray[long,ndim=2] traj_cell_2):

    cdef double D,d
    cdef int n1,n2,p1x,p1y,p2x,p2y,i,j
    cdef np.ndarray[long,ndim=1] p1,p2

    D=0
    n1=len(traj_cell_1)
    n2=len(traj_cell_2)

    for i from 0 <= i < n1:
        p1=traj_cell_1[i]
        p1x=p1[0]
        p1y=p1[1]
        d=INFINITY
        for j from 0 <= j < n2:
            p2=traj_cell_2[j]
            p2x=p2[0]
            p2y=p2[1]
            d=fmin(d,c_eucl_dist(p1x,p1y,p2x,p2y))
        D = D + d
    D = D/n1
    return D

def c_owd_grid_brut(np.ndarray[long,ndim=2] traj_cell_1, np.ndarray[long,ndim=2] traj_cell_2):

    cdef double D,d
    cdef int n1,n2,p1x,p1y,p2x,p2y,i,j
    cdef np.ndarray[long,ndim=1] p1,p2

    D=0
    n1=len(traj_cell_1)
    n2=len(traj_cell_2)

    for i from 0 <= i < n1:
        p1=traj_cell_1[i]
        p1x=p1[0]
        p1y=p1[1]
        d=INFINITY
        for j from 0 <= j < n2:
            p2=traj_cell_2[j]
            p2x=p2[0]
            p2y=p2[1]
            d=fmin(d,c_eucl_dist(p1x,p1y,p2x,p2y))
        D = D + d
    D = D/n1
    return D


cdef double _sowd_grid_brut(np.ndarray[long,ndim=2] traj_cell_1,np.ndarray[long,ndim=2] traj_cell_2):

    cdef double sowd_brut_dist

    sowd_brut_dist = _owd_grid_brut(traj_cell_1,traj_cell_2)+_owd_grid_brut(traj_cell_2,traj_cell_1)
    return sowd_brut_dist/2

def c_sowd_grid_brut(np.ndarray[long,ndim=2] traj_cell_1,np.ndarray[long,ndim=2] traj_cell_2):

    cdef double sowd_brut_dist

    sowd_brut_dist = _owd_grid_brut(traj_cell_1,traj_cell_2)+_owd_grid_brut(traj_cell_2,traj_cell_1)
    return sowd_brut_dist/2



cdef np.ndarray[long,ndim=1] _find_first_min_points( np.ndarray[np.float64_t,ndim=1] pt, int n):

    cdef np.ndarray[np.uint8_t,ndim=1,cast=True] min_points
    cdef np.ndarray[long,ndim=1] min_points_index
    cdef double pti,ptip,ptim
    cdef bool m_p,ciip,cimi

    min_points = np.empty([n],dtype=bool)

    if n == 1:
        min_points[0] = True
    else:
        pti=pt[0]
        ptip=pt[1]
        ciip=  pti < ptip
        min_points[0] = ciip
        for i from 1 <= i < n-1:
            pti = pt[i]
            ptip = pt[i+1]
            ptim = pt[i-1]
            ciip = pti < ptip
            cimi = pti < ptim
            m_p =  (ciip and cimi)
            min_points [i] = m_p
        ptim = pt[n - 2]
        pti = pt[n - 1]
        cimi =  pti < ptim
        min_points[n-1] = cimi

    min_points_index = np.where(np.array(min_points))[0]
    return min_points_index


def c_find_first_min_points( np.ndarray[np.float64_t,ndim=1] pt, int n):

    cdef np.ndarray[np.uint8_t,ndim=1,cast=True] min_points
    cdef np.ndarray[long,ndim=1] min_points_index
    cdef double pti,ptip,ptim
    cdef bool m_p,ciip,cimi

    min_points = np.empty([n],dtype=bool)

    if n == 1:
        min_points[0] = True
    else:
        pti=pt[0]
        ptip=pt[1]
        ciip=  pti < ptip
        min_points[0] = ciip
        for i from 1 <= i < n-1:
            pti = pt[i]
            ptip = pt[i+1]
            ptim = pt[i-1]
            ciip = pti < ptip
            cimi = pti < ptim
            m_p =  (ciip and cimi)
            min_points [i] = m_p
        ptim = pt[n - 2]
        pti = pt[n - 1]
        cimi =  pti < ptim
        min_points[n-1] = cimi

    min_points_index = np.where(np.array(min_points))[0]
    return min_points_index


def c_owd_grid( np.ndarray[long,ndim=2] traj_cell_1,np.ndarray[long,ndim=2] traj_cell_2):

    cdef int n1,n2,px,py,p_precx,p_precy,p2x,p2y,i,j,n_S_old,ig,pgx,pgy,rmin,rmax,pgpx,pgpy,pgpmx,pgpmy,pgppx,pgppy
    cdef double dpp2,D,d,dist,dist_back,dist_forw
    cdef np.ndarray[long,ndim=1] p,p_prec,pg,p2,S_old,pgp,pgpp,pgpm,S_ind
    cdef np.ndarray[np.float64_t,ndim=1] p_t2

    n1 = len(traj_cell_1)
    n2 = len(traj_cell_2)

    p = traj_cell_1[0]
    px = p[0]
    py = p[1]

    p2 = traj_cell_2[0]
    p2x=p2[0]
    p2y=p2[1]

    p_t2 = np.empty([n2],dtype=float)
    dpp2=c_eucl_dist(px,py,p2x,p2y)
    p_t2[0] = dpp2
    for i from 1 <= i < n2:
        p2 = traj_cell_2[i]
        p2x=p2[0]
        p2y=p2[1]
        dpp2_=c_eucl_dist(px,py,p2x,p2y)
        dpp2=fmin(dpp2,dpp2_)
        p_t2[i] = dpp2_

    S_old = _find_first_min_points(p_t2, n2)
    D = dpp2
    for i from 1 <= i < n1:
        p_prec = p
        p_precx=px
        p_precy=py
        p = traj_cell_1[i]
        px=p[0]
        py=p[1]
        S_ind = np.zeros([n2],dtype=int)
        d = INFINITY
        n_S_old=len(S_old)
        for j from 0 <= j < n_S_old:
            ig = S_old[j]
            pg = traj_cell_2[ig]
            pgx=pg[0]
            pgy=pg[1]
            if (p_precy == py and pgx != p_precx) or (p_precx == px and pgy != p_precy):
                S_ind[ig] = 1
                dppg=c_eucl_dist(px,py,pgx,pgy)
                d=fmin(d,dppg)
            else:
                if j == 0:
                    if n_S_old == 1 :
                        rmin = 0
                        rmax = n2
                    else:
                        rmin = 0
                        rmax = S_old[j+1]
                elif j == n_S_old - 1:
                    rmin = S_old[j-1]
                    rmax = n2
                else:
                    rmin=S_old[j-1]+1
                    rmax= S_old[j+1]
                for igp from rmin <= igp < rmax:
                    pgp = traj_cell_2[igp]
                    pgpx = pgp[0]
                    pgpy = pgp[1]
                    if (p_precy == py and pgpx == px ) or (p_precx == px and pgpy == py) or igp==ig  :
                        if igp !=0:
                            pgpm = traj_cell_2[igp-1]
                            pgpmx=pgpm[0]
                            pgpmy=pgpm[1]
                            dist_back=c_eucl_dist(pgpmx,pgpmy,px,py)
                        else:
                            dist_back = INFINITY
                        if igp !=n2-1:
                            pgpp = traj_cell_2[igp+1]
                            pgppx=pgpp[0]
                            pgppy=pgpp[1]
                            dist_forw=c_eucl_dist(pgppx,pgppy,px,py)
                        else:
                            dist_forw = INFINITY
                        dist = c_eucl_dist(pgpx,pgpy,px,py)
                        if dist < dist_back and dist < dist_forw:
                            if not S_ind[igp]:
                                S_ind[igp] = 1
                                d=fmin(d,dist)

        S_old=np.where(S_ind)[0]
        D += d
    return D/n1


cdef double _owd_grid( np.ndarray[long,ndim=2] traj_cell_1,np.ndarray[long,ndim=2] traj_cell_2):

    cdef int n1,n2,px,py,p_precx,p_precy,p2x,p2y,i,j,n_S_old,ig,pgx,pgy,rmin,rmax,pgpx,pgpy,pgpmx,pgpmy,pgppx,pgppy
    cdef double dpp2,D,d,dist,dist_back,dist_forw
    cdef np.ndarray[long,ndim=1] p,p_prec,pg,p2,S_old,pgp,pgpp,pgpm,S_ind
    cdef np.ndarray[np.float64_t,ndim=1] p_t2

    n1 = len(traj_cell_1)
    n2 = len(traj_cell_2)

    p = traj_cell_1[0]
    px = p[0]
    py = p[1]

    p2 = traj_cell_2[0]
    p2x=p2[0]
    p2y=p2[1]

    p_t2 = np.empty([n2],dtype=float)
    dpp2=c_eucl_dist(px,py,p2x,p2y)
    p_t2[0] = dpp2
    for i from 1 <= i < n2:
        p2 = traj_cell_2[i]
        p2x=p2[0]
        p2y=p2[1]
        dpp2_=c_eucl_dist(px,py,p2x,p2y)
        dpp2=fmin(dpp2,dpp2_)
        p_t2[i] = dpp2_

    S_old = _find_first_min_points(p_t2, n2)
    D = dpp2
    for i from 1 <= i < n1:
        p_prec = p
        p_precx=px
        p_precy=py
        p = traj_cell_1[i]
        px=p[0]
        py=p[1]
        S_ind = np.zeros([n2],dtype=int)
        d = INFINITY
        n_S_old=len(S_old)
        for j from 0 <= j < n_S_old:
            ig = S_old[j]
            pg = traj_cell_2[ig]
            pgx=pg[0]
            pgy=pg[1]
            if (p_precy == py and pgx != p_precx) or (p_precx == px and pgy != p_precy):
                S_ind[ig] = 1
                dppg=c_eucl_dist(px,py,pgx,pgy)
                d=fmin(d,dppg)
            else:
                if j == 0:
                    if n_S_old == 1 :
                        rmin = 0
                        rmax = n2
                    else:
                        rmin = 0
                        rmax = S_old[j+1]
                elif j == n_S_old - 1:
                    rmin = S_old[j-1]
                    rmax = n2
                else:
                    rmin=S_old[j-1]+1
                    rmax= S_old[j+1]
                for igp from rmin <= igp < rmax:
                    pgp = traj_cell_2[igp]
                    pgpx = pgp[0]
                    pgpy = pgp[1]
                    if (p_precy == py and pgpx == px ) or (p_precx == px and pgpy == py) or igp==ig  :
                        if igp !=0:
                            pgpm = traj_cell_2[igp-1]
                            pgpmx=pgpm[0]
                            pgpmy=pgpm[1]
                            dist_back=c_eucl_dist(pgpmx,pgpmy,px,py)
                        else:
                            dist_back = INFINITY
                        if igp !=n2-1:
                            pgpp = traj_cell_2[igp+1]
                            pgppx=pgpp[0]
                            pgppy=pgpp[1]
                            dist_forw=c_eucl_dist(pgppx,pgppy,px,py)
                        else:
                            dist_forw = INFINITY
                        dist = c_eucl_dist(pgpx,pgpy,px,py)
                        if dist < dist_back and dist < dist_forw:
                            if not S_ind[igp]:
                                S_ind[igp] = 1
                                d=fmin(d,dist)

        S_old=np.where(S_ind)[0]
        D += d
    return D/n1



cdef double _sowd_grid(np.ndarray[long,ndim=2] traj_cell_1,np.ndarray[long,ndim=2] traj_cell_2):

    cdef double sowd_dist

    sowd_dist = _owd_grid(traj_cell_1,traj_cell_2)+_owd_grid(traj_cell_2,traj_cell_1)
    return sowd_dist/2

def c_sowd_grid(np.ndarray[long,ndim=2] traj_cell_1,np.ndarray[long,ndim=2] traj_cell_2):

    cdef double sowd_dist

    sowd_dist = _owd_grid(traj_cell_1,traj_cell_2)+_owd_grid(traj_cell_2,traj_cell_1)
    return sowd_dist/2


STUFF = "Hi"

#cython: boundscheck=False
#cython: wraparound=False

from libc.math cimport fmin

from basic_euclidean import c_point_to_trajectory
from basic_geographical import c_point_to_path
cimport numpy as np
from numpy.math cimport INFINITY



################
## Euclidean ###
################

cdef double _e_spd (np.ndarray[np.float64_t,ndim=2] t1, np.ndarray[np.float64_t,ndim=2] t2):
    """
    Usage
    -----
    The spd-distance of trajectory t2 from trajectory t1
    The spd-distance is the sum of the all the point-to-trajectory distance of points of t1 from trajectory t2

    Parameters
    ----------
    param t1 :  len(t1)x2 numpy_array
    param t2 :  len(t2)x2 numpy_array

    Returns
    -------
    spd : float
           spd-distance of trajectory t2 from trajectory t1
    """
    cdef int nt,i
    cdef double spd,px,py

    nt=len(t1)
    spd=0
    for i from 0 <= i < (nt):
        px=t1[i,0]
        py=t1[i,1]
        spd=spd+c_point_to_trajectory(px,py,t2)
    spd=spd/nt
    return spd

def c_e_spd (np.ndarray[np.float64_t,ndim=2] t1, np.ndarray[np.float64_t,ndim=2] t2):
    """
    Usage
    -----
    The spd-distance of trajectory t2 from trajectory t1
    The spd-distance is the sum of the all the point-to-trajectory distance of points of t1 from trajectory t2

    Parameters
    ----------
    param t1 :  len(t1)x2 numpy_array
    param t2 :  len(t2)x2 numpy_array

    Returns
    -------
    spd : float
           spd-distance of trajectory t2 from trajectory t1
    """
    cdef int nt,i
    cdef double spd,px,py

    nt=len(t1)
    spd=0
    for i from 0 <= i < (nt):
        px=t1[i,0]
        py=t1[i,1]
        spd=spd+c_point_to_trajectory(px,py,t2)
    spd=spd/nt
    return spd

def c_e_sspd (np.ndarray[np.float64_t,ndim=2] t1, np.ndarray[np.float64_t,ndim=2] t2):
    """
    Usage
    -----
    The sspd-distance between trajectories t1 and t2.
    The sspd-distance is the mean of the spd-distance between of t1 from t2 and the spd-distance of t2 from t1.

    Parameters
    ----------
    param t1 :  len(t1)x2 numpy_array
    param t2 :  len(t2)x2 numpy_array

    Returns
    -------
    sspd : float
            sspd-distance of trajectory t2 from trajectory t1
    """
    cdef double sspd
    sspd=(_e_spd(t1,t2) + _e_spd(t2,t1))/2
    return sspd


###################
## Geographical ###
###################

cdef double _g_spd(np.ndarray[np.float64_t,ndim=2] t1, np.ndarray[np.float64_t,ndim=2] t2):
    """
    Usage
    -----
    The spd-distance of trajectory t2 from trajectory t1
    The spd-distance is the sum of the all the point-to-path distance of points of t1 from trajectory t2

    Parameters
    ----------
    param t1 :  len(t1)x2 numpy_array
    param t2 :  len(t2)x2 numpy_array

    Returns
    -------
    spd : float
           spd-distance of trajectory t2 from trajectory t1
    """
    cdef int n1,i,j
    cdef double dist,dist_j0
    cdef np.ndarray[np.float64_t,ndim=1] lats0,lons0,lats1,lons1

    n0=len(t1)
    n1=len(t2)
    lats0=t1[:,1]
    lons0=t1[:,0]
    lats1=t2[:,1]
    lons1=t2[:,0]
    dist=0
    for j from 0 <= j < n1  :
        dist_j0=INFINITY
        for i from 0 <= i < (n0-1):
            dist_j0=fmin(dist_j0,c_point_to_path(lons0[i],lats0[i],lons0[i+1],lats0[i+1],lons1[j],lats1[j]))
        dist=dist+dist_j0
    dist=float(dist)/n1
    return dist


def c_g_spd(np.ndarray[np.float64_t,ndim=2] t1, np.ndarray[np.float64_t,ndim=2] t2):
    """
    Usage
    -----
    The spd-distance of trajectory t2 from trajectory t1
    The spd-distance is the sum of the all the point-to-path distance of points of t1 from trajectory t2

    Parameters
    ----------
    param t1 :  len(t1)x2 numpy_array
    param t2 :  len(t2)x2 numpy_array

    Returns
    -------
    spd : float
           spd-distance of trajectory t2 from trajectory t1
    """
    cdef int n1,i,j
    cdef double dist,dist_j0
    cdef np.ndarray[np.float64_t,ndim=1] lats0,lons0,lats1,lons1

    n0=len(t1)
    n1=len(t2)
    lats0=t1[:,1]
    lons0=t1[:,0]
    lats1=t2[:,1]
    lons1=t2[:,0]
    dist=0
    for j from 0 <= j < n1:
        dist_j0=INFINITY
        for i from 0 <= i < (n0-1):
            dist_j0=fmin(dist_j0,c_point_to_path(lons0[i],lats0[i],lons0[i+1],lats0[i+1],lons1[j],lats1[j]))
        dist=dist+dist_j0
    dist=float(dist)/n1
    return dist


cdef double _g_sspd(np.ndarray[np.float64_t,ndim=2] t1, np.ndarray[np.float64_t,ndim=2] t2):
    """
    Usage
    -----
    The sspd-distance between trajectories t1 and t2.
    The sspd-distance is the mean of the spd-distance between of t1 from t2 and the spd-distance of t2 from t1.

    Parameters
    ----------
    param t1 :  len(t1)x2 numpy_array
    param t2 :  len(t2)x2 numpy_array


    Returns
    -------
    sspd : float
            sspd-distance of trajectory t2 from trajectory t1
    """

    cdef double dist

    dist=_g_spd(t1,t2) + _g_spd(t2,t1)
    return dist

def c_g_sspd(np.ndarray[np.float64_t,ndim=2] t1, np.ndarray[np.float64_t,ndim=2] t2):
    """
    Usage
    -----
    The sspd-distance between trajectories t1 and t2.
    The sspd-distance is the mean of the spd-distance between of t1 from t2 and the spd-distance of t2 from t1.

    Parameters
    ----------
    param t1 :  len(t1)x2 numpy_array
    param t2 :  len(t2)x2 numpy_array

    Returns
    -------
    sspd : float
            sspd-distance of trajectory t2 from trajectory t1
    """
    dist=_g_spd(t1,t2) + _g_spd(t2,t1)
    return dist
